import importlib
import abc
from typing import Dict, Any, Callable, Union, Optional, TypeVar, Generic, _GenericAlias
from functools import partial, wraps

import amino
from amino.util.string import snake_case
from amino.lazy import lazy

from amino.logging import amino_root_logger, Logging

A = TypeVar('A')
TC = TypeVar('TC', bound='TypeClass')


def normalize_type(tpe: type) -> type:
    return (
        tpe.__origin__
        if isinstance(tpe, _GenericAlias) else
        tpe
    )


class TypeClassMeta(type):

    def __new__(
            cls: type,
            name: str,
            bases: tuple,
            namespace: dict,
            tpe: type=None,
            pred: Callable=None,
            sub: type=None,
            **kw: dict
    ) -> type:
        inst = super().__new__(cls, name, bases, namespace, **kw)  # type: ignore
        if name != 'TypeClass':
            if tpe is not None:
                Instances.add_auto(tpe, inst())
            elif pred is not None:
                Instances.add_pred(pred, inst())
            elif sub is not None:
                Instances.add_pred(lambda t: issubclass(t, sub), inst())
        return inst

    def fatal(self, tpe: type) -> 'TypeClass':
        return Instances.lookup_fatal(self, tpe)

    lookup = fatal

    def fatal_for(self, a: Any) -> 'amino.maybe.Maybe[TypeClass]':
        return self.fatal(type(a))

    def m(self, tpe: type) -> 'amino.maybe.Maybe[TypeClass]':
        return amino.maybe.Maybe.check(Instances.lookup(self, tpe))

    def m_for(self, a: Any) -> 'amino.maybe.Maybe[TypeClass]':
        return self.m(type(a))

    def e(self, tpe: type) -> 'amino.maybe.Maybe[TypeClass]':
        normalized = normalize_type(tpe)
        return (
            self.m(normalized)
            .to_either(f'no `{self.__name__}` instance for {normalized} ({tpe})')
        )

    def e_for(self, a: A) -> 'amino.either.Either[str, TC]':
        return self.e(type(a))

    def exists_instance(self, tpe: type) -> bool:
        return self.m(tpe).present

    exists = exists_instance


class TypeClass(Logging, metaclass=TypeClassMeta):
    pass


class TypeClasses:

    @property
    def instances(self) -> Dict[type, TypeClass]:
        pass


class ImplicitInstancesNotFound(Exception):

    def __init__(self, mod: str, cls: str, name: str) -> None:
        msg = 'invalid implicit class path {}.{} for {}'.format(mod, cls, name)
        super().__init__(msg)


operators = (
    '__floordiv__',
    '__truediv__',
    '__mod__',
    '__or__',
    '__and__',
)


class ImplicitInstancesMeta(abc.ABCMeta):

    def __new__(cls: type, name: str, bases: tuple, namespace: dict, tpe: type=None, **kw: dict) -> type:
        inst = super().__new__(cls, name, bases, namespace,  # type: ignore
                               **kw)
        if name != 'ImplicitInstances' and tpe is not None:
            Instances.add_instances(tpe, inst())
        return inst


class ImplicitInstances(metaclass=ImplicitInstancesMeta):

    @lazy
    def instances(self) -> Dict[type, TypeClass]:
        return (
            TC.instances ** self._instances **  # type: ignore
            self._override_instances
        )

    @abc.abstractproperty
    def _instances(self) -> Dict[type, TypeClass]:
        ...

    @property
    def _override_instances(self) -> Dict[type, TypeClass]:
        from amino.map import Map
        return Map()


def lookup_implicit_instances(name: str, m: str, c: str) -> ImplicitInstances:
    err = ImplicitInstancesNotFound(m, c, name)
    try:
        mod = importlib.import_module(m)
    except ImportError:
        raise err
    else:
        if hasattr(mod, c):
            instances = getattr(mod, c)()
            return instances
        else:
            raise err


class InstancesMetadata:

    def __init__(self, name: str, mod: str, cls: str) -> None:
        self.name = name
        self.mod = mod
        self.cls = cls
        self._instances = None  # type: Optional[ImplicitInstances]

    def __str__(self) -> str:
        return '{}({}, {}, {})'.format(self.__class__.__name__, self.name,
                                       self.mod, self.cls)

    @property
    def instances(self) -> Dict[type, TypeClass]:
        if self._instances is None:
            self._instances = lookup_implicit_instances(self.name, self.mod, self.cls)
        return self._instances.instances


def infer_implicits_module(name: str) -> str:
    snake = snake_case(name)
    return 'amino.instances.{}'.format(snake)


def infer_implicits_class(name: str) -> str:
    return '{}Instances'.format(name)


class ImplicitsMeta(abc.ABCMeta):

    @staticmethod
    def _mk_operator(name: str) -> Callable:
        def dispatch(self: Any, other: Any) -> Any:
            return self._operator(name, other)
        return dispatch

    def _attach_operators(inst) -> None:
        for op in operators:
            setattr(inst, op, ImplicitsMeta._mk_operator(op))

    def __new__(cls: type, name: str, bases: tuple, namespace: dict, imp_mod: str=None, imp_cls: str=None,
                implicits: bool=False, auto: bool=False, **kw: Any) -> type:
        inst = super().__new__(cls, name, bases, namespace, **kw)  # type: ignore
        inst.auto = auto or getattr(inst, 'auto', False)
        if not implicits:
            return inst
        else:
            inst.instances_meta = None
            inst.name = name
            ImplicitsMeta._attach_operators(inst)
            if not inst.auto:
                if imp_mod is None:
                    imp_mod = infer_implicits_module(name)
                if imp_cls is None:
                    imp_cls = infer_implicits_class(name)
                meta = InstancesMetadata(name, imp_mod, imp_cls)
                inst.imp_mod = imp_mod
                inst.imp_cls = imp_cls
                inst.instances_meta = meta
                Instances.add(meta)
            return inst

    __copy__ = None


def tc_prop(f: Callable) -> Callable:
    f._tc_prop = True  # type: ignore
    return f


class ImplicitLookupError(Exception):
    pass


class Implicits(metaclass=ImplicitsMeta):
    permanent = True

    def _lookup_implicit_attr(self, name: str) -> Optional[Callable]:
        meta = type(self).instances_meta
        if meta is not None:
            return next((getattr(inst, name) for inst in meta.instances.v if hasattr(inst, name)), None)
        elif self.auto:
            return Instances.lookup_auto_attr(type(self), name)

    def _bound_implicit_attr(self, name: str) -> Union[None, Callable, partial]:
        f = self._lookup_implicit_attr(name)
        if f is not None:
            if hasattr(f, '_tc_prop'):
                return f(self)
            else:
                return partial(f, self)
        return None

    def _set_implicit_attr(self, name: str) -> Any:
        f = self._lookup_implicit_attr(name)
        if f is not None:
            @wraps(f)
            def wrap(self: Any, *a: Any, **kw: Any) -> Any:
                return f(self, *a, **kw)  # type: ignore
            g = property(wrap) if hasattr(f, '_tc_prop') else wrap
            setattr(type(self), name, g)
            return getattr(self, name)

    def __getattr__(self, name: str) -> Callable:
        try:
            imp = self._set_implicit_attr(name) if Implicits.permanent else self._bound_implicit_attr(name)
            if imp is None:
                raise AttributeError(f'`{self.__class__.__name__}` object has no attribute `{name}`')
            else:
                return imp
        except (RecursionError, ImplicitLookupError):
            raise ImplicitLookupError(f'recursion error for implicit attr `{name}` on `{self}`')

    def _operator(self, name: str, other: Any) -> Any:
        op = (self._set_implicit_attr(name) if Implicits.permanent else
              self._bound_implicit_attr(name))
        if op is None:
            err = '\'{}\' has no implicit operator \'{}\''.format(self, name)
            raise TypeError(err)
        else:
            return op(other)

    @property
    def dbg(self) -> Any:
        amino_root_logger.test(self)
        return self

    @property
    def dbgr(self) -> Any:
        v = self.log.verbose  # type: ignore
        v(repr(self))
        return self


class GlobalTypeClasses(TypeClasses):

    @property
    def instances(self) -> Dict[type, TypeClass]:
        from amino.map import Map
        from amino.tc.tap import Tap
        from amino.tc.show import Show
        return Map({Show: Show(), Tap: Tap()})


class F(Generic[A], Implicits):
    pass


TC = GlobalTypeClasses()


class ImplicitNotFound(Exception):

    def __init__(self, tc: type, tpe: type, msg: str=None) -> None:
        explanation = '' if msg is None else f': {msg}'
        msg = f'no type class found for {tc}[{tpe}]{explanation}'
        super().__init__(msg)


class AutoImplicitInstances(ImplicitInstances):

    def __init__(self, tpe: type, instance: TypeClass) -> None:
        self.tpe = tpe
        self.instance = instance

    @lazy
    def _instances(self) -> Dict[type, TypeClass]:
        from amino import Map
        return Map({self.tpe: self.instance})


class NoTypeClass(Exception):

    def __init__(self, inst: Any) -> None:
        super().__init__('invalid type class: {}'.format(inst))


class PredTypeClass:

    def __init__(self, pred: Callable[[type], bool], tc: TypeClass) -> None:
        self.pred = pred
        self.tc = tc


def unbounded_typevar(tc: type, v: TypeVar) -> None:
    raise ImplicitNotFound(tc, v, f'unbounded TypeVar cannot be used for typeclasses')


class AllInstances:

    def __init__(self) -> None:
        self.instances = dict()  # type: Dict[str, InstancesMetadata]
        self.auto_instances = dict()  # type: Dict[type, ImplicitInstances]
        self.pred_instances = (dict())  # type: Dict[type, amino.list.List[PredTypeClass]]

    def add(self, data: InstancesMetadata) -> None:
        self.instances[data.name] = data

    def add_instances(self, tpe: type, instances: ImplicitInstances) -> None:
        if tpe in self.auto_instances:
            self.auto_instances[tpe].instances.update(instances._instances)
        else:
            self.auto_instances[tpe] = instances

    def _tc_type(self, inst: TypeClass) -> type:
        mro = inst.__class__.__mro__
        if len(mro) < 2 or TypeClass not in mro:
            raise NoTypeClass(inst)
        return mro[1]

    def add_auto(self, tpe: type, inst: TypeClass) -> None:
        tc_type = self._tc_type(inst)
        self.add_instances(tpe, AutoImplicitInstances(tc_type, inst))

    def add_pred(self, pred: Callable[[type], bool], inst: TypeClass) -> None:
        import amino.list
        tc_type = self._tc_type(inst)
        data = PredTypeClass(pred, inst)
        if tc_type in self.pred_instances:
            self.pred_instances[tc_type].append(data)
        else:
            self.pred_instances[tc_type] = amino.list.List(data)

    def lookup_fatal(self, TC: type, G: type) -> TypeClass:
        tc = self.lookup(TC, G)
        if tc is None:
            raise ImplicitNotFound(TC, G)
        return tc

    def lookup(self, TC: type, G: type) -> Optional[TypeClass]:
        ''' Find an instance of the type class `TC` for type `G`.
        Iterates `G`'s parent classes, looking up instances for each,
        checking whether the instance is a subclass of the target type
        class `TC`.
        '''
        if isinstance(G, str):
            raise ImplicitNotFound(TC, G, f'{G} is a string annotation')
        if not isinstance(G, (type, TypeVar, _GenericAlias)):
            raise ImplicitNotFound(TC, G, f'{G} is neither type, _GenericAlias nor TypeVar: {type(G)}')
        match = lambda a: self._lookup_type(TC, a)
        def attach_type(tc: TypeClass) -> TypeClass:
            setattr(tc, 'tpe', G)
            return tc
        scrutinee = (
            (
                unbounded_typevar(TC, G)
                if G.__bound__ is None else
                G.__bound__
            )
            if isinstance(G, TypeVar) else
            G
        )
        safe_mro = lambda t: getattr(t, '__mro__', (t,))
        target = scrutinee.__origin__ if isinstance(scrutinee, _GenericAlias) else scrutinee
        mro = safe_mro(target)
        return next((attach_type(a) for a in map(match, mro) if a is not None), None)

    def lookup_auto_attr(self, tpe: type, name: str) -> Optional[Callable]:
        def check(t: type) -> Optional[Callable]:
            if t in self.auto_instances:
                ins = self.auto_instances[t]
                return next((getattr(inst, name) for inst in ins.instances.v if hasattr(inst, name)), None)
            else:
                return None
        return next((a for a in map(check, tpe.__mro__) if a is not None), None)

    def _lookup_type(self, TC: type, G: type) -> Optional[TypeClass]:
        auto = self._lookup_auto(TC, G)
        if auto is not None:
            return auto
        reg = self._lookup_regular(TC, G)
        if reg is not None:
            return reg
        return self._lookup_pred(TC, G)

    def _lookup_auto(self, TC: type, G: type) -> Optional[TypeClass]:
        if G in self.auto_instances:
            return next((b for a, b in self.auto_instances[G].instances.items() if isinstance(b, TC)), None)
        else:
            return None

    def _lookup_regular(self, TC: type, G: type) -> Optional[TypeClass]:
        if G.__name__ in self.instances:
            return next((b for a, b in self.instances[G.__name__].instances.items() if isinstance(b, TC)), None)
        else:
            return None

    def _lookup_pred(self, TC: type, G: type) -> Optional[TypeClass]:
        if TC in self.pred_instances:
            return next((a.tc for a in self.pred_instances[TC] if a.pred(G)), None)
        else:
            return None

Instances = AllInstances()  # type: AllInstances

__all__ = ('TypeClasses', 'TC', 'tc_prop', 'TypeClass')



import inspect
from typing import Tuple, Callable

from traceback import format_list, format_exception_only, FrameSummary, extract_tb

from amino.data.list import List, Lists
from amino.func import I


def sanitize_tb(tb: List[str]) -> List[str]:
    return tb.flat_map(lambda a: Lists.wrap(a.splitlines())) / (lambda a: a.rstrip())


def format_exception_error(exc: Exception) -> List[str]:
    return sanitize_tb(Lists.wrap(format_exception_only(type(exc), exc)))


def format_one_exception(
        exc: Exception,
        tb_filter: Callable[[List[FrameSummary]], List[FrameSummary]]=I,
        tb_formatter: Callable[[List[str]], List[str]]=I,
        exc_formatter: Callable[[List[str]], List[str]]=I,
) -> Tuple[List[str], List[str]]:
    e_str = exc_formatter(format_exception_error(exc))
    tb = tb_filter(Lists.wrap(extract_tb(exc.__traceback__)))
    tb_str = tb_formatter(sanitize_tb(Lists.wrap(format_list(tb))))
    return e_str, tb_str


def format_cause(exc: Exception, **kw) -> List[str]:
    from amino import Maybe
    return Maybe.optional(exc.__cause__) / (lambda a: format_exception(a, **kw)) / (lambda a: a.cons('Cause:')) | List()


def format_exception(exc: Exception, **kw) -> List[str]:
    e, tb = format_one_exception(exc, **kw)
    main = tb + e
    return main + format_cause(exc, **kw)


def format_stack(stack: List[inspect.FrameInfo]) -> List[str]:
    data = stack.map(lambda a: a[1:-2] + tuple(a[-2] or ['<string>']))
    return sanitize_tb(Lists.wrap(format_list(list(data))))


def format_current_stack() -> List[str]:
    return format_stack(Lists.wrap(inspect.stack()))


__all__ = ('sanitize_tb', 'format_exception', 'format_stack', 'format_current_stack')



from types import SimpleNamespace
from typing import Type, Generic, Any, TypeVar, Callable, Tuple, _GenericAlias, get_type_hints
import inspect

from amino.util.string import snake_case
from amino import ADT, Map, List, Lists, do, Do, Either, Try, Just, Nothing, Maybe
from amino.algebra import Algebra
from amino.logging import module_log

log = module_log()
Alg = TypeVar('Alg', bound=Algebra)
C = TypeVar('C', bound=Alg)
A = TypeVar('A')
B = TypeVar('B')


class CaseMeta(type):

    def __new__(cls: type, name: str, bases: tuple, namespace: SimpleNamespace, alg: Type[Alg]=None, **kw: Any) -> type:
        inst = super().__new__(cls, name, bases, namespace, **kw)
        if alg is not None:
            inst.case = case_dispatch(inst, alg)
        return inst


class Case(Generic[Alg, B], metaclass=CaseMeta):

    @classmethod
    def match(cls, scrutinee: Alg, *a: Any, **kw: Any) -> B:
        return cls().case(scrutinee, *a, **kw)

    def __call__(self, scrutinee: Alg, *a: Any, **kw: Any) -> B:
        return self.case(scrutinee, *a, **kw)


def normalize_type(tpe: Type[C]) -> Type[C]:
    return getattr(tpe, '__origin__', tpe)


def case_list(
        cls: Type[Case[C, B]],
        variants: List[Type[C]],
        alg: Type[C],
        has_default: bool,
) -> Map[Type[C], Callable[[Case[C, B]], B]]:
    @do(Maybe[Tuple[Type[C], Callable[[Case[C, B]], B]]])
    def is_handler(name: str, f: Callable) -> Do:
        effective = getattr(f, '__do_original', f)
        hints = yield Try(get_type_hints, effective).to_maybe
        spec = yield Try(inspect.getfullargspec, effective).to_maybe
        param_name = yield Lists.wrap(spec.args).lift(1)
        param_type = yield Map(hints).lift(param_name)
        yield (
            Just((normalize_type(param_type), f))
            if isinstance(param_type, type) and issubclass(param_type, alg) else
            Just((normalize_type(param_type), f))
            if isinstance(param_type, _GenericAlias) and issubclass(param_type.__origin__, alg) else
            Nothing
        )
    handlers = Lists.wrap(inspect.getmembers(cls, inspect.isfunction)).flat_map2(is_handler)
    def find_handler(variant: Type[C]) -> Callable[[Case[C, B]], B]:
        def not_found() -> None:
            if not has_default:
                raise Exception(f'no case defined for {variant} on {cls.__name__}')
        def match_handler(tpe: type, f: Callable) -> Maybe[Callable[[Case[C, B]], B]]:
            return Just(f) if issubclass(tpe, variant) else Nothing
        return handlers.find_map2(match_handler).get_or_else(not_found)
    return variants.map(find_handler)


# TODO determine default case from param type being the ADT
def case_dispatch(cls: Type[Case[C, B]], alg: Type[C]) -> Callable[[Case[C, B], C], B]:
    def error(func: Case[C, B], variant: Alg, *a: Any, **kw: Any) -> None:
        raise TypeError(f'no case defined for {variant} on {cls.__name__}')
    default = getattr(cls, 'case_default', error)
    has_default = default is not error
    cases = case_list(cls, alg.__algebra_variants__.sort_by(lambda a: a.__algebra_index__), alg, has_default)
    length = cases.length
    def case(func: Case[C, B], scrutinee: C, *a: Any, **kw: Any) -> B:
        index = getattr(scrutinee, '__algebra_index__', None)
        handler = (cases[index] or default) if index is not None and index < length else default
        return handler(func, scrutinee, *a, **kw)
    return case


class CaseRecMeta(type):

    def __new__(cls, name: str, bases: tuple, namespace: dict, alg: Type[Alg]=None, **kw: Any) -> type:
        inst = super().__new__(cls, name, bases, namespace, **kw)
        if alg is not None:
            inst.case = case_dispatch(inst, alg)
        return inst


class CaseRec(Generic[Alg, A], metaclass=CaseRecMeta):

    def __call__(self, scrutinee: Alg, *a: Any, **kw: Any) -> 'Rec[Alg, A]':
        return Rec(self, scrutinee, a, kw)


class RecStep(Generic[Alg, A], ADT['RecStep[Alg, A]']):
    pass


class Rec(Generic[Alg, A], RecStep[Alg, A]):

    def __init__(self, func: CaseRec[Alg, A], scrutinee: Alg, args: list, kwargs: dict) -> None:
        self.func = func
        self.scrutinee = scrutinee
        self.args = args
        self.kwargs = kwargs

    def eval(self) -> A:
        return eval_case_rec(self)


class Term(Generic[Alg, A], RecStep[Alg, A]):

    def __init__(self, result: A) -> None:
        self.result = result


def eval_case_rec(step: Rec[Alg, A]) -> A:
    while True:
        step = step.func.case(step.scrutinee, *step.args, **step.kwargs)
        if isinstance(step, Term):
            return step.result
        elif not isinstance(step, Rec):
            raise Exception(f'invalid result in CaseRec step: {step}')


__all__ = ('Case', 'CaseRec', 'Term')



'''Stolen from https://github.com/prashnts/hues
'''

from typing import Union, Any, Callable, Type
from collections import namedtuple
from functools import reduce, partial


ANSIColors = namedtuple(
    'ANSIColors',
    [
        'black',
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan',
        'white',
    ]
)

ANSIStyles = namedtuple(
    'ANSIStyles',
    [
        'reset',
        'bold',
        'italic',
        'underline',
        'defaultfg',
        'defaultbg',
    ]
)

# Style Codes
STYLE = ANSIStyles(0, 1, 3, 4, 39, 49)

# Regular Colors
FG = ANSIColors(*range(30, 38))
BG = ANSIColors(*range(40, 48))

# High Intensity Colors
HI_FG = ANSIColors(*range(90, 98))
HI_BG = ANSIColors(*range(100, 108))

# Terminal sequence format
SEQ = '\033[%sm'


def gen_keywords(*args: Union[ANSIColors, ANSIStyles], **kwargs: Union[ANSIColors, ANSIStyles]) -> tuple:
    '''generate single escape sequence mapping.'''
    fields: tuple = tuple()
    values: tuple = tuple()
    for tpl in args:
        fields += tpl._fields
        values += tpl
    for prefix, tpl in kwargs.items():
        fields += tuple(map(lambda x: '_'.join([prefix, x]), tpl._fields))
        values += tpl
    return namedtuple('ANSISequences', fields)(*values)

KEYWORDS = gen_keywords(STYLE, FG, bg=BG, bright=HI_FG, bg_bright=HI_BG)


def zero_break(stack: tuple) -> tuple:
    '''Handle Resets in input stack.
    Breaks the input stack if a Reset operator (zero) is encountered.
    '''
    reducer = lambda x, y: tuple() if y == 0 else x + (y,)
    return reduce(reducer, stack, tuple())


def annihilate(predicate: tuple, stack: tuple) -> tuple:
    '''Squash and reduce the input stack.
    Removes the elements of input that match predicate and only keeps the last
    match at the end of the stack.
    '''
    extra = tuple(filter(lambda x: x not in predicate, stack))
    head = reduce(lambda x, y: y if y in predicate else x, stack, None)
    return extra + (head,) if head else extra


def annihilator(predicate: tuple) -> Callable[[tuple], tuple]:
    '''Build a partial annihilator for given predicate.'''
    return partial(annihilate, predicate)


def dedup(stack: tuple) -> tuple:
    '''Remove duplicates from the stack in first-seen order.'''
    # Initializes with an accumulator and then reduces the stack with first match
    # deduplication.
    reducer = lambda x, y: x if y in x else x + (y,)
    return reduce(reducer, stack, tuple())


def apply(funcs: tuple, stack: tuple) -> tuple:
    '''Apply functions to the stack, passing the resulting stack to next state.'''
    return reduce(lambda x, y: y(x), funcs, stack)


OPTIMIZATION_STEPS = (
    zero_break,               # Handle Resets using `reset`.
    annihilator(FG + HI_FG),  # Squash foreground colors to the last value.
    annihilator(BG + HI_BG),  # Squash background colors to the last value.
    dedup,                    # Remove duplicates in (remaining) style values.
)
optimize = partial(apply, OPTIMIZATION_STEPS)


def colorize(string: str, stack: tuple) -> str:
    '''Apply optimal ANSI escape sequences to the string.'''
    codes = optimize(stack)
    if len(codes):
        prefix = SEQ % ';'.join(map(str, codes))
        suffix = SEQ % STYLE.reset
        return prefix + string + suffix
    else:
        return string


class HueString(str):

    def __new__(cls, string: str, hue_stack: tuple=None) -> Type['HueString']:
        return super(HueString, cls).__new__(cls)

    def __init__(self, string: str, hue_stack: tuple=tuple()) -> None:
        self.__string = string
        self.__hue_stack = hue_stack

    def __getattr__(self, attr: str) -> 'HueString':
        try:
            code = getattr(KEYWORDS, attr)
            hues = self.__hue_stack + (code,)
            return HueString(self.__string, hue_stack=hues)
        except AttributeError as e:
            raise e

    @property
    def colorized(self) -> str:
        return colorize(self.__string, self.__hue_stack)


def huestr(s: str) -> HueString:
    return HueString(s)


def col(a: Any, c: Callable[[HueString], HueString]) -> str:
    return c(huestr(str(a))).colorized


def red(a: Any) -> str:
    return col(a, lambda a: a.red)


def green(a: Any) -> str:
    return col(a, lambda a: a.green)


def yellow(a: Any) -> str:
    return col(a, lambda a: a.yellow)


def blue(a: Any) -> str:
    return col(a, lambda a: a.blue)


def cyan(a: Any) -> str:
    return col(a, lambda a: a.cyan)


def magenta(a: Any) -> str:
    return col(a, lambda a: a.magenta)


__all__ = ('huestr', 'red', 'green', 'yellow', 'blue', 'cyan', 'magenta')



import abc
from typing import Generic, TypeVar, Tuple
from traceback import FrameSummary

from amino import ADT, Either, Right, Left, Dat, Nil
from amino.util.trace import cframe
from amino.state import State
from amino.compute.exception import ComputeException

A = TypeVar('A')
B = TypeVar('B')


class CResult(Generic[A], ADT['CResult[A]']):

    @abc.abstractproperty
    def to_either(self) -> Either[Exception, A]:
        ...


class CSuccess(Generic[A], CResult[A]):

    def __init__(self, value: A) -> None:
        self.value = value

    @property
    def to_either(self) -> Either[Exception, A]:
        return Right(self.value)


class CError(Generic[A], CResult[A]):

    def __init__(self, error: str) -> None:
        self.error = error

    @property
    def to_either(self) -> Either[Exception, A]:
        return Left(Exception(self.error))


class CFatal(Generic[A], CResult[A]):

    def __init__(self, exception: Exception) -> None:
        self.exception = exception

    @property
    def to_either(self) -> Either[Exception, A]:
        return Left(self.exception)


class Thunk(Generic[A, B], Dat['Thunk[A, B]']):

    @staticmethod
    def cons(thunk: State[A, B], frame: FrameSummary=None) -> 'Thunk[A, B]':
        return Thunk(thunk, frame or cframe())

    def __init__(self, thunk: State[A, B], frame: FrameSummary) -> None:
        self.thunk = thunk
        self.frame = frame


def eval_thunk(resource: A, thunk: Thunk[A, B]) -> Tuple[A, B]:
    try:
        return thunk.thunk.run(resource).value
    except ComputeException as e:
        raise e
    except Exception as e:
        raise ComputeException('', Nil, e, thunk.frame)


__all__ = ('CResult', 'NSuccess', 'NError', 'NFatal', 'Thunk', 'eval_thunk')



from numbers import Number
from typing import Any, Callable

from amino import may, Right, Left, Maybe, Either


@may
def try_convert_int(data: Any) -> Maybe[int]:
    if (
            isinstance(data, Number) or
            (isinstance(data, str) and data.isdigit())
    ):
        return int(data)


def parse_int(i: Any) -> Either[str, int]:
    return Right(i) if isinstance(i, int) else (
        Right(int(i)) if isinstance(i, str) and i.isdigit() else
        Left('could not parse int {}'.format(i))
    )


def add(inc: int) -> Callable[[int], int]:
    def add(z: int) -> int:
        return z + inc
    return add


def sub(inc: int) -> Callable[[int], int]:
    def sub(z: int) -> int:
        return z - inc
    return sub


__all__ = ('try_convert_int', 'parse_int', 'add',)



import inspect
from typing import TypeVar, Callable, Any, Generic, Type
from functools import reduce

from amino import Map, Boolean
from amino.case import Case
from amino.algebra import Algebra

A = TypeVar('A')
B = TypeVar('B')
X = TypeVar('X')
Alg = TypeVar('Alg', bound=Algebra)


class Bindings:

    def __init__(self, bindings: Map[str, type]) -> None:
        self.bindings = bindings


def function_bindings_relay(f: Callable) -> Boolean:
    target = f.__do_original if hasattr(f, '__do') else f
    anno = inspect.getfullargspec(target).annotations
    return Boolean(anno.get('bindings', None) is Bindings)


def case_bindings_relay(pm: Type[Case[Alg, B]]) -> Boolean:
    return function_bindings_relay(pm.__init__)


def check_bound(rep: str, t: Type[A], b: Type[B], bindings: Map) -> None:
    binding = bindings.lift(t).get_or_fail(lambda: f'no binding for {t} => {b} in {rep}')
    b.m(binding).get_or_fail(lambda: f'no instance of {t} => {b} for {binding} in {rep}')


def check_bounds(rep: str, bounds: Map, bindings: Map) -> None:
    for t, b in bounds.items():
        check_bound(rep, t, b, bindings)


class BoundedFunction(Generic[A]):

    def __init__(self, f: Callable, bindings: Bindings, relay: Boolean) -> None:
        self.f = f
        self.bindings = bindings
        self.relay = relay

    def __call__(self, *a, **kw) -> Callable:
        return self.f(self.bindings, *a, **kw) if self.relay else self.f(*a, **kw)


class ContextBound(Generic[A]):

    def __init__(self, f: Callable, bounds: Map) -> None:
        self.f = f
        self.bounds = bounds
        self.relay = function_bindings_relay(f)

    def __call__(self, *a: Bindings, **kw) -> Callable:
        bindings = reduce(lambda a, b: a ** b.bindings, a, Map(kw))
        check_bounds(str(self.f), self.bounds, bindings)
        return BoundedFunction(self.f, Bindings(bindings), self.relay)

    def call(self, *a, **kw) -> A:
        return self.f(*a, **kw)


class BoundedCase(Generic[Alg, B]):

    def __init__(self, pm: Case[Alg, B], bindings: Bindings, relay: Boolean) -> None:
        self.pm = pm
        self.bindings = bindings
        self.relay = relay

    def __call__(self, *a, **kw) -> Callable:
        return self.pm(self.bindings, *a, **kw) if self.relay else self.pm(*a, **kw)


class CaseContextBound(Generic[Alg, B]):

    def __init__(self, pm: Type[Case[Alg, B]], bounds: Map) -> None:
        self.pm = pm
        self.bounds = bounds
        self.relay = case_bindings_relay(pm)

    def __call__(self, *a: Bindings, **kw: type) -> Case[Alg, B]:
        bindings = reduce(lambda a, b: a ** b.bindings, a, Map(kw))
        check_bounds(self.pm.__name__, self.bounds, bindings)
        return BoundedCase(self.pm, Bindings(bindings), self.relay)

    def call(self, *a, **kw) -> A:
        return self.pm(*a, **kw)


def context(*bindings: ContextBound, **tpes: Any) -> Callable:
    def dec(f: Callable) -> Callable:
        return (
            CaseContextBound(f, Map(tpes))
            if isinstance(f, type) and issubclass(f, Case) else
            ContextBound(f, Map(tpes))
        )
    return dec

__all__ = ('context',)



from typing import TypeVar, Callable, Any, Type, Tuple

from amino import Map
from amino.lazy import lazy
from amino.tc.functor import Functor
from amino.tc.base import ImplicitInstances, F
from amino.tc.traverse import Traverse
from amino.tc.monoid import Monoid
from amino.tc.monad import Monad

A = TypeVar('A', covariant=True)
B = TypeVar('B')
C = TypeVar('C')
G = TypeVar('G', bound=F)


class MapInstances(ImplicitInstances):

    @lazy
    def _instances(self):
        from amino import Map
        return Map(
            {
                Functor: MapFunctor(),
                Traverse: MapTraverse(),
                Monoid: MapMonoid(),
            }
        )


class MapFunctor(Functor):

    def map(self, fa: Map[Any, A], f: Callable[[A], B]) -> Map[Any, B]:
        return fa.valmap(f)


class MapTraverse(Traverse):

    def traverse(self, fa: Map[Any, A], f: Callable[[A], B], tpe: Type[G]) -> G:
        monad = Monad.fatal(tpe)
        def folder(z, kv: Tuple[A, B]):
            k, v = kv
            return monad.map2(z.product(f(v)), lambda l, b: l.cat((k, b)))
        return fa.to_list.fold_left(monad.pure(Map()))(folder)


class MapMonoid(Monoid):

    @property
    def empty(self):
        return Map()

    def combine(self, fa: Map, fb: Map):
        return fa ** fb

__all__ = ('MapInstances',)



import itertools
from functools import wraps
from typing import Generic, TypeVar, Callable, Union, Any, Iterable, Generator

from toolz import concatv

from amino.data.list import List
from amino.maybe import Just, Empty, Maybe
from amino.tc.base import Implicits
from amino.func import I
from amino.util.string import safe_string

A = TypeVar('A')
B = TypeVar('B')


class LazyList(Generic[A], Implicits, implicits=True):
    _default_chunk_size = 20

    def fetch(f: Callable) -> Callable:
        @wraps(f)
        def wrapper(self: 'LazyList', index: Union[slice, int]) -> Any:
            if isinstance(index, int) and index < 0:
                self._drain()
            else:
                self._fetch(index)
            return f(self, index)
        return wrapper

    def __init__(self, source: Iterable, init: List[A]=List(), chunk_size: Union[int, None]=None) -> None:
        self.source = iter(source)
        self.strict = init
        self._chunk_size = chunk_size or self._default_chunk_size

    @fetch
    def __getitem__(self, index: int) -> Maybe[A]:
        return self.strict.__getitem__(index)

    def __len__(self) -> int:
        return self.length

    @property
    def length(self) -> int:
        return self.drain.length

    @property
    def _one(self) -> Generator:
        try:
            yield next(self.source)
        except StopIteration:
            pass

    def _fetch(self, index: Union[slice, int, float]) -> None:
        count = index.stop if isinstance(index, slice) else index + 1
        def chunk() -> List[A]:
            for i in range(self._chunk_size):
                yield from self._one
        def gen() -> None:
            while True:
                if self.strict.length < count:
                    c = list(chunk())
                    self.strict = self.strict + c
                    if len(c) == self._chunk_size:
                        continue
                break
        gen()

    @property
    def drain(self) -> List[A]:
        return self._drain()

    def _drain(self) -> List[A]:
        self._fetch(float('inf'))
        return self.strict

    def copy(self, wrap_source: Callable, trans_strict: Callable[[List[A]], List[B]]) -> 'LazyList[B]':
        a, b = itertools.tee(self.source)
        self.source = a
        return LazyList(wrap_source(b), trans_strict(self.strict), self._chunk_size)

    @fetch
    def lift(self, index: int) -> Maybe[A]:
        return self.strict.lift(index)

    @property
    def head(self) -> Maybe[A]:
        return self.lift(0)

    @property
    def last(self) -> Maybe[A]:
        return self.lift(-1)

    def _drain_find(self, abort: Callable[[A], bool]) -> Maybe[A]:
        culprit = Empty()
        def gen() -> Generator:
            nonlocal culprit
            while True:
                try:
                    el = next(self.source)
                    yield el
                    if abort(el):
                        culprit = Just(el)
                        break
                except StopIteration:
                    break
        drained = List.wrap(list(gen()))
        self.strict = self.strict + drained
        return culprit

    def foreach(self, f: Callable[[A], None]) -> None:
        self.drain.foreach(f)

    @fetch
    def min_length(self, index: int) -> bool:
        return self.strict.length >= index

    @fetch
    def max_length(self, index: int) -> bool:
        return self.strict.length <= index

    @property
    def empty(self) -> bool:
        return self.max_length(0)

    def append(self, other: 'LazyList[A]') -> 'LazyList[A]':
        return self.copy(lambda s: concatv(s, self.strict, other.source, other.strict), lambda s: List())

    __add__ = append

    def __repr__(self) -> str:
        strict = (self.strict / safe_string).mk_string(', ')
        return '{}({} {!r})'.format(self.__class__.__name__, strict,
                                    self.source)

    def mk_string(self, sep: str='') -> str:
        return sep.join((self / str).drain)

    @property
    def join_lines(self) -> str:
        return self.mk_string('\n')

    @property
    def join_comma(self) -> str:
        return self.mk_string(',')

    def cons(self, a: A) -> 'LazyList[A]':
        return self.copy(I, lambda s: s.cons(a))

    def cat(self, a: A) -> 'LazyList[A]':
        return self.copy(lambda xs: itertools.chain(xs, (a,)), I)

    def collect(self, f: Callable[[A], Maybe[B]]) -> 'LazyList[B]':
        return self.copy(lambda s: (a._unsafe_value for a in map(f, s) if a.present), lambda a: a.collect(f))


def lazy_list(f: Callable) -> Callable:
    @wraps(f)
    def w(*a: Any, **kw: Any) -> LazyList:
        return LazyList(f(*a, **kw))
    return w


def lazy_list_prop(f: Callable) -> property:
    return property(lazy_list(f))


class LazyLists:

    @staticmethod
    def cons(*a: A) -> LazyList[A]:
        return LazyList(a)

    @staticmethod
    def empty() -> LazyList[A]:
        return LazyLists.cons()

    @staticmethod
    def range(*rng: int) -> LazyList[int]:
        return LazyList(range(*rng))

__all__ = ('LazyList', 'lazy_list', 'LazyLists')



from typing import Generic, TypeVar, Callable, Tuple, cast, Type, Any

from lenses import UnboundLens

from amino.tc.base import ImplicitsMeta, Implicits
from amino.tc.monad import Monad
from amino.tc.zip import Zip
from amino.instances.list import ListTraverse
from amino import List, curried
from amino.util.string import ToStr
from amino.state.base import StateT
from amino.either import Either



A = TypeVar('A')
B = TypeVar('B')
S = TypeVar('S')
R = TypeVar('R')
ST1 = TypeVar('ST1')
E = TypeVar('E')

monad: Monad = cast(Monad, Monad.fatal(Either))


class EitherStateCtor(Generic[S]):

    def inspect(self, f: Callable[[S], A]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return EitherState.apply(g)

    def inspect_f(self, f: Callable[[S], Either[E, A]]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return EitherState.apply(g)

    def pure(self, a: A) -> 'EitherState[E, S, A]':
        return EitherState.apply(lambda s: monad.pure((s, a)))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'EitherState[E, S, A]':
        return EitherState.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: Either[E, A]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return EitherState.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'EitherState[E, S, None]':
        return EitherState.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], Either[E, S]]) -> 'EitherState[E, S, None]':
        return EitherState.apply(lambda s: f(s).map(lambda a: (a, None)))

    def get(self) -> 'EitherState[E, S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'EitherState[E, S, None]':
        return EitherState.pure(None)



class EitherStateMeta(ImplicitsMeta):

    def cons(self, run_f: Either[E, Callable[[S], Either[E, Tuple[S, A]]]]) -> 'EitherState[E, S, A]':
        return self(run_f)

    def apply(self, f: Callable[[S], Either[E, Tuple[S, A]]]) -> 'EitherState[E, S, A]':
        return self.cons(monad.pure(f))

    def apply_f(self, run_f: Either[E, Callable[[S], Either[E, Tuple[S, A]]]]) -> 'EitherState[E, S, A]':
        return self.cons(run_f)

    def inspect(self, f: Callable[[S], A]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return monad.pure((s, f(s)))
        return self.apply(g)

    def inspect_f(self, f: Callable[[S], Either[E, A]]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return f(s).map(lambda a: (s, a))
        return self.apply(g)

    def pure(self, a: A) -> 'EitherState[E, S, A]':
        return self.apply(lambda s: monad.pure((s, a)))

    def reset(self, s: S, a: A) -> 'EitherState[E, S, A]':
        return self.apply(lambda _: monad.pure((s, a)))

    def reset_t(self, t: Tuple[S, A]) -> 'EitherState[E, S, A]':
        return self.apply(lambda _: monad.pure(t))

    def delay(self, fa: Callable[..., A], *a: Any, **kw: Any) -> 'EitherState[E, S, A]':
        return self.apply(lambda s: monad.pure((s, fa(*a, **kw))))

    def lift(self, fa: Either[E, A]) -> 'EitherState[E, S, A]':
        def g(s: S) -> Either[E, Tuple[S, A]]:
            return fa.map(lambda a: (s, a))
        return self.apply(g)

    def modify(self, f: Callable[[S], S]) -> 'EitherState[E, S, None]':
        return self.apply(lambda s: monad.pure((f(s), None)))

    def modify_f(self, f: Callable[[S], Either[E, S]]) -> 'EitherState[E, S, None]':
        return self.apply(lambda s: f(s).map(lambda a: (a, None)))

    def set(self, s: S) -> 'EitherState[E, S, None]':
        return self.modify(lambda s0: s)

    def get(self) -> 'EitherState[E, S, S]':
        return self.inspect(lambda a: a)

    @property
    def unit(self) -> 'EitherState[E, S, None]':
        return EitherState.pure(None)

    def s(self, tpe: Type[S]) -> EitherStateCtor[S]:
        return EitherStateCtor()



class EitherState(Generic[E, S, A], StateT, ToStr, Implicits, implicits=True, auto=True, metaclass=EitherStateMeta):

    def __init__(self, run_f: Either[E, Callable[[S], Either[E, Tuple[S, A]]]]) -> None:
        self.run_f = run_f

    @property
    def cls(self) -> Type['EitherState[E, S, A]']:
        return type(self)

    def run(self, s: S) -> Either[E, Tuple[S, A]]:
        return self.run_f.flat_map(lambda f: f(s))

    def run_s(self, s: S) -> Either[E, S]:
        return self.run(s).map(lambda a: a[0])

    def run_a(self, s: S) -> Either[E, S]:
        return self.run(s).map(lambda a: a[1])

    def _arg_desc(self) -> List[str]:
        return List(str(self.run_f))

    def flat_map_f(self, f: Callable[[A], Either[E, B]]) -> 'EitherState[E, S, B]':
        def h(s: S, a: A) -> Either[E, Tuple[S, B]]:
            return f(a).map(lambda b: (s, b))
        def g(fsa: Either[E, Tuple[S, A]]) -> Either[E, Tuple[S, B]]:
            return fsa.flat_map2(h)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform(self, f: Callable[[Tuple[S, A]], Tuple[S, B]]) -> 'EitherState[E, S, B]':
        def g(fsa: Either[E, Tuple[S, A]]) -> Either[E, Tuple[S, B]]:
            return fsa.map2(f)
        run_f1 = self.run_f.map(lambda sfsa: lambda a: g(sfsa(a)))
        return self.cls.apply_f(run_f1)

    def transform_s(self, f: Callable[[R], S], g: Callable[[R, S], R]) -> 'EitherState[E, R, A]':
        def trans(sfsa: Callable[[S], Either[E, Tuple[S, A]]], r: R) -> Either[E, Tuple[R, A]]:
            s = f(r)
            return sfsa(s).map2(lambda s, a: (g(r, s), a))
        return self.cls.apply_f(self.run_f.map(curried(trans)))

    def transform_f(self, tpe: Type[ST1], f: Callable[[Either[E, Tuple[S, A]]], Any]) -> ST1:
        def trans(s: S) -> Any:
            return f(self.run(s))
        return tpe.apply(trans)  # type: ignore

    def zoom(self, l: UnboundLens) -> 'EitherState[E, R, A]':
        return self.transform_s(l.get(), lambda r, s: l.set(s)(r))

    transform_s_lens = zoom

    def read_zoom(self, l: UnboundLens) -> 'EitherState[E, R, A]':
        return self.transform_s(l.get(), lambda r, s: r)

    transform_s_lens_read = read_zoom

    def flat_map(self, f: Callable[[A], 'EitherState[E, S, B]']) -> 'EitherState[E, S, B]':
        return Monad_EitherState.flat_map(self, f)


def run_function(s: EitherState[E, S, A]) -> Either[E, Callable[[S], Either[E, Tuple[S, A]]]]:
    try:
        return s.run_f
    except Exception as e:
        if not isinstance(s, EitherState):
            raise TypeError(f'flatMapped {s} into EitherState')
        else:
            raise


class EitherStateMonad(Monad, tpe=EitherState):

    def pure(self, a: A) -> EitherState[E, S, A]:  # type: ignore
        return EitherState.pure(a)

    def flat_map(  # type: ignore
            self,
            fa: EitherState[E, S, A],
            f: Callable[[A], EitherState[E, S, B]]
    ) -> EitherState[E, S, B]:
        def h(s: S, a: A) -> Either[E, Tuple[S, B]]:
            return f(a).run(s)
        def g(fsa: Either[E, Tuple[S, A]]) -> Either[E, Tuple[S, B]]:
            return fsa.flat_map2(h)
        def i(sfsa: Callable[[S], Either[E, Tuple[S, A]]]) -> Callable[[S], Either[E, Tuple[S, B]]]:
            return lambda a: g(sfsa(a))
        run_f1 = run_function(fa).map(i)
        return EitherState.apply_f(run_f1)


Monad_EitherState = EitherStateMonad()


class EitherStateZip(Zip, tpe=EitherState):

    def zip(
            self,
            fa: EitherState[E, S, A],
            fb: EitherState[E, S, A],
            *fs: EitherState[E, S, A],
    ) -> EitherState[E, S, List[A]]:
        v = ListTraverse().sequence(List(fa, fb, *fs), EitherState)  # type: ignore
        return cast(EitherState[E, S, List[A]], v)



__all__ = ('EitherState',)
